@@ -8,7 +8,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2021	PgPool Global Development Group
+ * Copyright (c) 2003-2025	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -97,19 +97,17 @@ static char *pwdfMatchesString(char *buf, char *token);
 PCPConnInfo *
 pcp_connect(char *hostname, int port, char *username, char *password, FILE *Pfdebug)
 {
-	struct sockaddr_in addr;
 	struct sockaddr_un unix_addr;
-	struct hostent *hp;
 	char	   *password_from_file = NULL;
 	char		os_user[256];
 	PCPConnInfo *pcpConn = palloc0(sizeof(PCPConnInfo));
 	int			fd;
 	int			on = 1;
-	int			len;
 
 	pcpConn->connState = PCP_CONNECTION_NOT_CONNECTED;
 	pcpConn->Pfdebug = Pfdebug;
 
+	/* Unix domain socket? */
 	if (hostname == NULL || *hostname == '\0' || *hostname == '/')
 	{
 		char	   *path;
@@ -153,49 +151,83 @@ pcp_connect(char *hostname, int port, char *username, char *password, FILE *Pfde
 	}
 	else
 	{
-		fd = socket(AF_INET, SOCK_STREAM, 0);
-		if (fd < 0)
+		/* Inet domain socket case */
+		char	   *portstr;
+		int			ret;
+		struct addrinfo *res;
+		struct addrinfo *walk;
+		struct addrinfo hints;
+
+		/*
+		 * getaddrinfo() requires a string because it also accepts service names,
+		 * such as "http".
+		 */
+		if (asprintf(&portstr, "%d", port) == -1)
 		{
 			pcp_internal_error(pcpConn,
-							   "ERROR: failed to create INET domain socket with error \"%s\"", strerror(errno));
+							   "ERROR: asprintf() failed");
 			pcpConn->connState = PCP_CONNECTION_BAD;
 			return pcpConn;
 		}
 
-		if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
-					   (char *) &on, sizeof(on)) < 0)
-		{
-			close(fd);
+		memset(&hints, 0, sizeof(struct addrinfo));
+		hints.ai_family = PF_UNSPEC;
+		hints.ai_socktype = SOCK_STREAM;
 
+		if ((ret = getaddrinfo(hostname, portstr, &hints, &res)) != 0)
+		{
 			pcp_internal_error(pcpConn,
-							   "ERROR: set socket option failed with error \"%s\"", strerror(errno));
+							   "ERROR: getaddrinfo failed \"%s\"",
+							   gai_strerror(ret));
 			pcpConn->connState = PCP_CONNECTION_BAD;
+			free(portstr);
 			return pcpConn;
 		}
 
-		memset((char *) &addr, 0, sizeof(addr));
-		addr.sin_family = AF_INET;
-		hp = gethostbyname(hostname);
-		if ((hp == NULL) || (hp->h_addrtype != AF_INET))
-		{
-			close(fd);
-			pcp_internal_error(pcpConn,
-							   "ERROR: could not retrieve hostname. gethostbyname failed with error \"%s\"", strerror(errno));
-			pcpConn->connState = PCP_CONNECTION_BAD;
-			return pcpConn;
+		free(portstr);
 
+		fd = -1;
+		for (walk = res; walk != NULL; walk = walk->ai_next)
+		{
+			fd = socket(walk->ai_family, walk->ai_socktype, walk->ai_protocol);
+			if (fd < 0)
+			{
+				pcp_internal_error(pcpConn,
+								   "ERROR: failed to create INET domain socket with error \"%s\"",
+								   strerror(errno));
+				freeaddrinfo(res);
+				return pcpConn;
+			}
+
+			if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
+						   (char *) &on, sizeof(on)) < 0)
+			{
+				close(fd);
+				freeaddrinfo(res);
+				pcp_internal_error(pcpConn,
+								   "ERROR: set socket option failed with error \"%s\"", strerror(errno));
+				pcpConn->connState = PCP_CONNECTION_BAD;
+				return pcpConn;
+			}
+
+			if (connect(fd, walk->ai_addr, walk->ai_addrlen) < 0)
+			{
+				close(fd);
+				freeaddrinfo(res);
+				pcp_internal_error(pcpConn,
+								   "ERROR: connection to host \"%s\" failed with error \"%s\"", hostname, strerror(errno));
+				pcpConn->connState = PCP_CONNECTION_BAD;
+				return pcpConn;
+			}
+			break;	/* successfully connected */
 		}
-		memmove((char *) &(addr.sin_addr),
-				(char *) hp->h_addr,
-				hp->h_length);
-		addr.sin_port = htons(port);
 
-		len = sizeof(struct sockaddr_in);
-		if (connect(fd, (struct sockaddr *) &addr, len) < 0)
+		freeaddrinfo(res);
+		/* no address available */
+		if (fd == -1)
 		{
-			close(fd);
 			pcp_internal_error(pcpConn,
-							   "ERROR: connection to host \"%s\" failed with error \"%s\"", hostname, strerror(errno));
+							   "ERROR: connection to host \"%s\" failed", hostname);
 			pcpConn->connState = PCP_CONNECTION_BAD;
 			return pcpConn;
 		}
@@ -210,6 +242,7 @@ pcp_connect(char *hostname, int port, char *username, char *password, FILE *Pfde
 		pcpConn->connState = PCP_CONNECTION_BAD;
 		return pcpConn;
 	}
+
 	pcpConn->connState = PCP_CONNECTION_CONNECTED;
 
 	/*
@@ -390,6 +423,94 @@ static PCPResultInfo * process_pcp_response(PCPConnInfo * pcpConn, char sentMsg)
 
 		switch (toc)
 		{
+			case 'a':			/* set configuration parameter */
+				if (sentMsg != 'A')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_command_complete_response(pcpConn, buf, rsize);
+				break;
+
+			case 'b':			/* status request */
+				if (sentMsg != 'B')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_pool_status_response(pcpConn, buf, rsize);
+				break;
+
+			case 'c':			/* attach node */
+				if (sentMsg != 'C' && sentMsg != 'O')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_command_complete_response(pcpConn, buf, rsize);
+				break;
+
+			case 'd':			/* detach node */
+				if (sentMsg != 'D' && sentMsg != 'J')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_command_complete_response(pcpConn, buf, rsize);
+				break;
+
+			case 'E':			/* error */
+				setResultStatus(pcpConn, PCP_RES_BACKEND_ERROR);
+				process_error_response(pcpConn, toc, buf);
+				break;
+
+			case 'g':			/* invalidate query cache */
+				if (sentMsg != 'G')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_command_complete_response(pcpConn, buf, rsize);
+				break;
+
+			case 'h':			/* health check stats */
+				if (sentMsg != 'H')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_health_check_stats_response(pcpConn, buf, rsize);
+				break;
+
+			case 'i':			/* node info */
+				if (sentMsg != 'I')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_node_info_response(pcpConn, buf, rsize);
+				break;
+
+			case 'l':			/* node count */
+				if (sentMsg != 'L')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_pcp_node_count_response(pcpConn, buf, rsize);
+				break;
+
+			case 'm':			/* salt info response */
+				if (sentMsg != 'M')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_salt_info_response(pcpConn, buf, rsize);
+				break;
+
+			case 'N':			/* error response */
+				process_error_response(pcpConn, toc, buf);
+				pfree(buf);
+				continue;
+				break;
+
+			case 'n':			/* process count */
+				if (sentMsg != 'N')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_process_count_response(pcpConn, buf, rsize);
+				break;
+
+			case 'p':			/* process info */
+				if (sentMsg != 'P')
+					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
+				else
+					process_process_info_response(pcpConn, buf, rsize);
+				break;
+
 			case 'r':			/* Authentication Response */
 				{
 					if (sentMsg != 'R')
@@ -409,106 +530,33 @@ static PCPResultInfo * process_pcp_response(PCPConnInfo * pcpConn, char sentMsg)
 					}
 				}
 				break;
-			case 'm':
-				if (sentMsg != 'M')
+
+			case 't':			/* shutdown request */
+				if (sentMsg != 'T')
 					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
 				else
-					process_salt_info_response(pcpConn, buf, rsize);
+					setResultStatus(pcpConn, PCP_RES_COMMAND_OK);
 				break;
 
-			case 'E':
-				setResultStatus(pcpConn, PCP_RES_BACKEND_ERROR);
-				process_error_response(pcpConn, toc, buf);
-				break;
-
-			case 'N':
-				process_error_response(pcpConn, toc, buf);
-				pfree(buf);
-				continue;
-				break;
-
-			case 'i':
-				if (sentMsg != 'I')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_node_info_response(pcpConn, buf, rsize);
-				break;
-
-			case 'h':
-				if (sentMsg != 'H')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_health_check_stats_response(pcpConn, buf, rsize);
-				break;
-
-			case 'l':
-				if (sentMsg != 'L')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_pcp_node_count_response(pcpConn, buf, rsize);
-				break;
-
-			case 'c':
-				if (sentMsg != 'C' && sentMsg != 'O')
+			case 'v':			/* pcp_log_rotate */
+				if (sentMsg != 'V')
 					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
 				else
 					process_command_complete_response(pcpConn, buf, rsize);
 				break;
 
-			case 'd':
-				if (sentMsg != 'D' && sentMsg != 'J')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_command_complete_response(pcpConn, buf, rsize);
-				break;
-
-			case 'a':
-				if (sentMsg != 'A')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_command_complete_response(pcpConn, buf, rsize);
-				break;
-
-			case 'z':
-				if (sentMsg != 'Z')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_command_complete_response(pcpConn, buf, rsize);
-				break;
-
-			case 'w':
+			case 'w':			/* watchdog info */
 				if (sentMsg != 'W')
 					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
 				else
 					process_watchdog_info_response(pcpConn, buf, rsize);
 				break;
 
-			case 'p':
-				if (sentMsg != 'P')
+			case 'z':			/* command complete */
+				if (sentMsg != 'Z')
 					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
 				else
-					process_process_info_response(pcpConn, buf, rsize);
-				break;
-
-			case 'n':
-				if (sentMsg != 'N')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_process_count_response(pcpConn, buf, rsize);
-				break;
-
-			case 'b':
-				if (sentMsg != 'B')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					process_pool_status_response(pcpConn, buf, rsize);
-				break;
-
-			case 't':
-				if (sentMsg != 'T')
-					setResultStatus(pcpConn, PCP_RES_BAD_RESPONSE);
-				else
-					setResultStatus(pcpConn, PCP_RES_COMMAND_OK);
+					process_command_complete_response(pcpConn, buf, rsize);
 				break;
 
 			default:
@@ -770,8 +818,16 @@ process_node_info_response(PCPConnInfo * pcpConn, char *buf, int len)
 		if (index == NULL)
 			goto INVALID_RESPONSE;
 		index++;
+
 		strlcpy(backend_info->pg_role, index, sizeof(backend_info->pg_role));
 
+		index = (char *) memchr(index, '\0', len);
+		if (index == NULL)
+			goto INVALID_RESPONSE;
+
+		index++;
+		backend_info->standby_delay_by_time = atol(index);
+
 		index = (char *) memchr(index, '\0', len);
 		if (index == NULL)
 			goto INVALID_RESPONSE;
@@ -918,6 +974,53 @@ pcp_reload_config(PCPConnInfo * pcpConn,char command_scope)
 	return process_pcp_response(pcpConn, 'Z');
 }
 
+PCPResultInfo *
+pcp_log_rotate(PCPConnInfo * pcpConn,char command_scope)
+{
+	int                     wsize;
+/*
+ * pcp packet format for pcp_log_rotate
+ * v[size][command_scope]
+ */
+	if (PCPConnectionStatus(pcpConn) != PCP_CONNECTION_OK)
+	{
+	   pcp_internal_error(pcpConn, "invalid PCP connection");
+	   return NULL;
+	}
+
+	pcp_write(pcpConn->pcpConn, "V", 1);
+	wsize = htonl(sizeof(int) + sizeof(char));
+	pcp_write(pcpConn->pcpConn, &wsize, sizeof(int));
+	pcp_write(pcpConn->pcpConn, &command_scope, sizeof(char));
+	if (PCPFlush(pcpConn) < 0)
+	   return NULL;
+	if (pcpConn->Pfdebug)
+	   fprintf(pcpConn->Pfdebug, "DEBUG: send: tos=\"Z\", len=%d\n", ntohl(wsize));
+
+	return process_pcp_response(pcpConn, 'V');
+}
+
+PCPResultInfo *
+pcp_invalidate_query_cache(PCPConnInfo * pcpConn)
+{
+	int                     wsize;
+
+	if (PCPConnectionStatus(pcpConn) != PCP_CONNECTION_OK)
+	{
+	   pcp_internal_error(pcpConn, "invalid PCP connection");
+	   return NULL;
+	}
+
+	pcp_write(pcpConn->pcpConn, "G", 1);
+	wsize = htonl(sizeof(int));
+	pcp_write(pcpConn->pcpConn, &wsize, sizeof(int));
+	if (PCPFlush(pcpConn) < 0)
+	   return NULL;
+	if (pcpConn->Pfdebug)
+	   fprintf(pcpConn->Pfdebug, "DEBUG: send: tos=\"G\", len=%d\n", ntohl(wsize));
+
+	return process_pcp_response(pcpConn, 'G');
+}
 
 /*
  * Process health check response from PCP server.
